#include "PCurveFunc.h"
#include <math.h>


ParametricCurveTracer * GetParametricCurveTracer(tjs_int type)
{
	ParametricCurveTracer * curve = NULL;
	switch (type) {
	case PCURVE_TYPE_STUB:
		curve = new StubCurveTracer();
		break;
	case PCURVE_TYPE_KAG3_SPLINE:
		curve = new KAG3SplineTracer();
		break;
	case PCURVE_TYPE_CUBIC_SPLINE_OPEN:
		curve = new CubicSplineTracerOpen();
		break;
	case PCURVE_TYPE_CUBIC_SPLINE_CLOSE:
		curve = new CubicSplineTracerClose();
		break;
	case PCURVE_TYPE_BEZIER_OPEN:
		curve = new BezierTracerOpen();
		break;
	case PCURVE_TYPE_LAGRANGE_INTERPOLATION:
		curve = new LagrangeCurveTracer();
		break;
	case PCURVE_TYPE_CHAIKIN_OPEN:
		curve = new ChaikinTracerOpen();
		break;
	case PCURVE_TYPE_CHAIKIN_OPEN_PST:
		curve = new ChaikinTracerOpenPST();
		break;
	case PCURVE_TYPE_CHAIKIN_CLOSE:
		curve = new ChaikinTracerClose();
		break;

	default:
		break;
	}

	return curve;
}

void FreeParametricCurveTracer(ParametricCurveTracer * curve)
{
	if (curve) delete curve;
}


// C^[tF[X
ParametricCurveTracer::ParametricCurveTracer()
{
	xpath = ypath = NULL;
	numPoints = accuracy = numNodes = 0;
	inactivate();
}

ParametricCurveTracer::~ParametricCurveTracer()
{
}




//--------------------------------
// Stub
//--------------------------------

void StubCurveTracer::init(tjs_int * xp, tjs_int * yp, tjs_int n, tjs_int ac)
{
	clear();
	if (n >= getRequiredPoints() && xp && yp) {
		this->numPoints = n;
		this->xpath = xp;
		this->ypath = yp;
		this->accuracy = 1;
		activate();
	}
	return;
}

void StubCurveTracer::clear()
{
	xpath = ypath = NULL;
	inactivate();
}



//--------------------------------
// KAG3 Impl.
//--------------------------------
void KAG3SplineTracer::init(tjs_int * xp, tjs_int * yp, tjs_int n, tjs_int ac)
{
	clear();
	if (n >= getRequiredPoints() && ac > 0 && xp && yp) {
		this->numPoints = n;
		this->xpath = xp;
		this->ypath = yp;
		this->accuracy = ac;
		makeTable();
		activate();
	}

	return;
}

void KAG3SplineTracer::clear()
{
	releaseMem();
	xpath = ypath = NULL;
	inactivate();
}



// XvCԂɕKvȃ[NvZ
bool KAG3SplineTracer::makeTable()
{
	tjs_int N = this->numPoints;

	allocMem();
	tjs_real * tx = this->zx;
	tjs_real * ty = this->zy;
	tjs_int * x = this->xpath;
	tjs_int * y = this->ypath;

	tx[0] = 0;
	ty[0] = 0;
	tx[N - 1] = 0;
	ty[N - 1] = 0;

	tjs_real * tmpx = new tjs_real[N];
	tjs_real * tmpy = new tjs_real[N];

	for (tjs_int i = N - 2; i >= 0; i--) {
		tmpx[i + 1] = x[i + 1] - x[i];
		tmpy[i + 1] = y[i + 1] - y[i];
	}

	tx[1] = tmpx[2] - tmpx[1] - tx[0];
	ty[1] = tmpy[2] - tmpy[1] - ty[0];

	tmpx[1] = 4.0;
	tmpy[1] = 4.0;

	for (tjs_int i = 1; i < N - 2; i++) {
		tjs_real ax = 1.0 / tmpx[i];
		tjs_real ay = 1.0 / tmpy[i];
		tx[i + 1] = tmpx[i + 2] - tmpx[i + 1] - tmpx[i] * ax;
		ty[i + 1] = tmpy[i + 2] - tmpy[i + 1] - tmpy[i] * ay;
		tmpx[i + 1] = 4.0 - ax;
		tmpy[i + 1] = 4.0 - ay;
	}

	tx[N - 2] -= tx[N - 1];
	ty[N - 2] -= ty[N - 1];

	for (tjs_int i = N - 2; i > 0; i--) {
		tx[i] = (tx[i] - tx[i + 1]) / tmpx[i];
		ty[i] = (ty[i] - ty[i + 1]) / tmpy[i];
	}

	delete[] tmpx;
	delete[] tmpy;

	return true;
}


bool KAG3SplineTracer::getCoord(tjs_int index, tjs_int subindex, tjs_real & cx, tjs_real & cy)
{
	tjs_int dindex = index * accuracy + subindex;
	index = dindex / accuracy;
	if (index >= numPoints) return false;
	subindex = dindex % accuracy;
	tjs_real ratio = (tjs_real)subindex / (tjs_real)accuracy;
	tjs_real * tx = this->zx;
	tjs_real * ty = this->zy;
	tjs_int * xp = this->xpath;
	tjs_int * yp = this->ypath;

	cx = (((tx[index + 1] - tx[index]) * ratio +
		tx[index] * 3.0) * ratio +
		((tjs_real)(xp[index + 1] - xp[index]) -
		(tx[index] * 2.0 + tx[index + 1]))) * ratio +
		xp[index];

	cy = (((ty[index + 1] - ty[index]) * ratio +
		ty[index] * 3.0) * ratio +
		((tjs_real)(yp[index + 1] - yp[index]) -
		(ty[index] * 2.0 + ty[index + 1]))) * ratio +
		(tjs_real)yp[index];

	return true;
}


void KAG3SplineTracer::releaseMem()
{
	DELETE_ARRAY(zx)
	DELETE_ARRAY(zy)
}

void KAG3SplineTracer::allocMem()
{
	releaseMem();
	this->zx = new tjs_real[numPoints];
	this->zy = new tjs_real[numPoints];
}




//--------------------------------
// Cubic Spline (Open)
//--------------------------------
void CubicSplineTracerOpen::init(tjs_int * xp, tjs_int * yp, tjs_int n, tjs_int ac)
{
	clear();
	if (n >= getRequiredPoints() && ac > 0 && xp && yp) {
		this->numPoints = n;
		this->xpath = xp;
		this->ypath = yp;
		this->accuracy = ac;
		this->numNodes = (numPoints - 1) * accuracy;
		makeTable();
		activate();
	}

	return;
}

void CubicSplineTracerOpen::clear()
{
	releaseMem();
	xpath = ypath = NULL;
	inactivate();
}



void CubicSplineTracerOpen::makeTable()
{
	allocMem();

	tjs_int * x = this->xpath;
	tjs_int * y = this->ypath;
	tjs_real * p = this->pz;
	tjs_int N = this->numPoints;

	p[0] = 0;

	for (tjs_int i = 1; i < N; i++) {
		tjs_int t1 = x[i] - x[i - 1];
		tjs_int t2 = y[i] - y[i - 1];
		p[i] = p[i - 1] + sqrt(t1*t1 + t2*t2);
	}

	for (tjs_int i = 1; i < N; i++) {
		p[i] /= p[N - 1];
	}

	internalMakeTable(p, x, za);
	internalMakeTable(p, y, zb);
	createNodeIndex();
}

void CubicSplineTracerOpen::internalMakeTable(tjs_real * x, tjs_int * y, tjs_real * z)
{
	tjs_int N = numPoints;
	tjs_real * h = new tjs_real[N];
	tjs_real * d = new tjs_real[N];

	z[0] = 0;
	z[N - 1] = 0;	/* [_ł y''(x) */
    for (tjs_int i = 0; i < N - 1; i++) {
		h[i    ] =  x[i + 1] - x[i];
		d[i + 1] = (y[i + 1] - y[i]) / h[i];
	}
	z[1] = d[2] - d[1] - h[0] * z[0];
	d[1] = 2.0 * (x[2] - x[0]);
	for (tjs_int i = 1; i < N - 2; i++) {
		tjs_int t = h[i] / d[i];
		z[i + 1] = d[i + 2] - d[i + 1] - z[i] * t;
		d[i + 1] = 2.0 * (x[i + 2] - x[i]) - h[i] * t;
	}
	z[N - 2] -= h[N - 2] * z[N - 1];
	for (tjs_int i = N - 2; i > 0; i--) {
		z[i] = (z[i] - h[i] * z[i + 1]) / d[i];
	}

	delete[] h;
	delete[] d;
}


bool CubicSplineTracerOpen::getCoord(tjs_int dindex, tjs_real & cx, tjs_real & cy)
{
	bool rv = false;
	if (0 <= dindex && dindex <= this->numNodes) {
		cx = internalGetCoord(dindex, this->xpath, this->za);
		cy = internalGetCoord(dindex, this->ypath, this->zb);
		rv = true;
	}
	return rv;
}

tjs_real CubicSplineTracerOpen::internalGetCoord(tjs_int index, tjs_int * y, tjs_real * z)
{
	tjs_real * p = this->pz;
	tjs_real t = (tjs_real)index / (tjs_real)numNodes;
	tjs_int i = this->nodeIndex[index];

	if (i > 0) i--;

	tjs_real h = p[i + 1] - p[i];
	tjs_real d = t - p[i];

	return (((z[i + 1] - z[i]) * d / h + z[i] * 3) * d
		+ ((tjs_real)(y[i + 1] - y[i]) / h
		- (z[i] * 2 + z[i + 1]) * h)) * d + (tjs_real)y[i];
}

void CubicSplineTracerOpen::createNodeIndex()
{
	tjs_real * p = this->pz;
	tjs_int * ni = this->nodeIndex;
	tjs_int N = this->numPoints;
	tjs_int ac = this->numNodes;

	tjs_int k = 0;

	for (tjs_int i = 0; i <= ac; i++) {
		tjs_real t = (tjs_real)i / (tjs_real)ac;
		if (t > p[k] && (k < N - 1)) {
			k++;
		}
		ni[i] = k;
	}
}

void CubicSplineTracerOpen::releaseMem()
{
	DELETE_ARRAY(za)
	DELETE_ARRAY(zb)
	DELETE_ARRAY(pz)
	DELETE_ARRAY(nodeIndex)
}

void CubicSplineTracerOpen::allocMem()
{
	releaseMem();
	this->pz = new tjs_real[numPoints];
	this->za = new tjs_real[numPoints];
	this->zb = new tjs_real[numPoints];
	this->nodeIndex = new tjs_int[numNodes + 1];
}





//--------------------------------
// Cubic Spline (Close)
//--------------------------------
void CubicSplineTracerClose::init(tjs_int * xp, tjs_int * yp, tjs_int n, tjs_int ac)
{
	clear();
	if (n >= getRequiredPoints() && ac > 0 && xp && yp && (xp[0] == xp[n-1]) && (yp[0] == yp[n-1])) {
		this->numPoints = n;
		this->xpath = xp;
		this->ypath = yp;
		this->accuracy = ac;
		this->numNodes = (numPoints - 1) * accuracy;
		makeTable();
		activate();
	}

	return;
}


void CubicSplineTracerClose::internalMakeTable(tjs_real * x, tjs_int * y, tjs_real * z)
{
	tjs_int N = numPoints - 1;
	tjs_real * h = new tjs_real[N + 1];
	tjs_real * d = new tjs_real[N + 1];
	tjs_real * w = new tjs_real[N + 1];

	for (tjs_int i = 0; i < N; i++) {
		h[i] = x[i + 1] - x[i];
		w[i] = (y[i + 1] - y[i]) / h[i];
	}
	w[N] = w[0];

	for (tjs_int i = 1; i < N; i++) {
		d[i] = 2.0 * (x[i + 1] - x[i - 1]);
	}
	d[N] = 2.0 * (h[N - 1] + h[0]);

	for (tjs_int i = 1; i <= N; i++) {
		z[i] = w[i] - w[i - 1];
	}

	w[1] = h[0];
	w[N - 1] = h[N - 1];
	w[N] = d[N];
	for (tjs_int i = 2; i < N - 1; i++) {
		w[i] = 0.0;
	}
	for (tjs_int i = 1; i < N; i++) {
		tjs_real t = h[i] / d[i];
		z[i + 1] = z[i + 1] - z[i] * t;
		d[i + 1] = d[i + 1] - h[i] * t;
		w[i + 1] = w[i + 1] - w[i] * t;
	}

	w[0] = w[N];
	z[0] = z[N];
	for (tjs_int i = N - 2; i >= 0; i--) {
		tjs_real t = h[i] / d[i + 1];
		z[i] = z[i] - z[i + 1] * t;
		w[i] = w[i] - w[i + 1] * t;
	}

	tjs_real t = z[0] / w[0];
	z[0] = t;
	z[N] = t;
	for (tjs_int i = 1; i < N; i++) {
		z[i] = (z[i] - w[i] * t) / d[i];
	}

	delete[] h;
	delete[] d;
	delete[] w;
}







//--------------------------------
// Bezier curve
//--------------------------------
void BezierTracerOpen::init(tjs_int * xp, tjs_int * yp, tjs_int n, tjs_int ac)
{
	clear();
	if (n >= getRequiredPoints() && ac > 0 && xp && yp) {
		this->numPoints = n;
		this->xpath = xp;
		this->ypath = yp;
		this->accuracy = ac;
		this->numNodes = (numPoints - 1) * accuracy;
		activate();
	}

	return;
}

void BezierTracerOpen::clear()
{
	xpath = ypath = NULL;
	inactivate();
}



// protected
// static
// gݍ킹 nCr ̈ꗗԂBvf n+1B
// generated by bezier.pl
tjs_real * BezierTracerOpen::getCombinationCoeffs(tjs_int n)
{
	const tjs_int nMax = 16;
	static tjs_real * coeffsLarge = NULL;
	static tjs_real coeffs_00[1] = { 1 };
	static tjs_real coeffs_01[2] = { 1, 1 };
	static tjs_real coeffs_02[3] = { 1, 2, 1 };
	static tjs_real coeffs_03[4] = { 1, 3, 3, 1 };
	static tjs_real coeffs_04[5] = { 1, 4, 6, 4, 1 };
	static tjs_real coeffs_05[6] = { 1, 5, 10, 10, 5, 1 };
	static tjs_real coeffs_06[7] = { 1, 6, 15, 20, 15, 6, 1 };
	static tjs_real coeffs_07[8] = { 1, 7, 21, 35, 35, 21, 7, 1 };
	static tjs_real coeffs_08[9] = { 1, 8, 28, 56, 70, 56, 28, 8, 1 };
	static tjs_real coeffs_09[10] = { 1, 9, 36, 84, 126, 126, 84, 36, 9, 1 };
	static tjs_real coeffs_10[11] = { 1, 10, 45, 120, 210, 252, 210, 120, 45, 10, 1 };
	static tjs_real coeffs_11[12] = { 1, 11, 55, 165, 330, 462, 462, 330, 165, 55, 11, 1 };
	static tjs_real coeffs_12[13] = { 1, 12, 66, 220, 495, 792, 924, 792, 495, 220, 66, 12, 1 };
	static tjs_real coeffs_13[14] = { 1, 13, 78, 286, 715, 1287, 1716, 1716, 1287, 715, 286, 78, 13, 1 };
	static tjs_real coeffs_14[15] = { 1, 14, 91, 364, 1001, 2002, 3003, 3432, 3003, 2002, 1001, 364, 91, 14, 1 };
	static tjs_real coeffs_15[16] = { 1, 15, 105, 455, 1365, 3003, 5005, 6435, 6435, 5005, 3003, 1365, 455, 105, 15, 1 };
	static tjs_real coeffs_16[17] = { 1, 16, 120, 560, 1820, 4368, 8008, 11440, 12870, 11440, 8008, 4368, 1820, 560, 120, 16, 1 };
	static tjs_real *coeffs[nMax+1] = { coeffs_00, coeffs_01, coeffs_02, coeffs_03, coeffs_04, coeffs_05, coeffs_06, coeffs_07, coeffs_08, coeffs_09, coeffs_10, coeffs_11, coeffs_12, coeffs_13, coeffs_14, coeffs_15, coeffs_16 };
	tjs_real * rv = NULL;
	if (n > 0 && n <= nMax) {
		rv = coeffs[n];
	} else if (n > nMax) {
		if (coeffsLarge) delete coeffsLarge;
		coeffsLarge = new tjs_real[n+1];
		tjs_real * prevCoeffs = coeffs[nMax];
		coeffsLarge[0] = coeffsLarge[n] = 1;
		for (tjs_int i = 1; i < n; i++) {
			coeffsLarge[i] = getCombinationCoeffOverX(n, i, nMax + 1, prevCoeffs[i]);
		}
	}

	return rv;
}

// protected
// 傫Ȑ̑gݍ킹 nCr ԂB
// QgčċAIɎZoB
tjs_real BezierTracerOpen::getCombinationCoeffOverX(
	tjs_int n, 
	tjs_int r, 
	tjs_int next_n, 
	tjs_int old_nCr)
{
	if (next_n > n) return old_nCr;

	tjs_int new_nCr = (r < next_n && r > 0) ? (old_nCr * next_n / (next_n - r)) : 1;
	tjs_int rv = getCombinationCoeffOverX(n, r, next_n + 1, new_nCr);

	return rv;
}



bool BezierTracerOpen::getCoord(tjs_int dindex, tjs_real & cx, tjs_real & cy)
{
	if (dindex < 0 || dindex > numNodes) return false;

	bool rv = false;

	tjs_real t = (tjs_real)dindex / (tjs_real)numNodes;
	tjs_real s = 1.0 - t;
	tjs_int * xp = this->xpath;
	tjs_int * yp = this->ypath;

	tjs_real x = 0;
	tjs_real y = 0;

	tjs_int N = numPoints - 1;
	tjs_real * combination = getCombinationCoeffs(N);

	if (combination) {
		tjs_real k;

		k = pow(s, N);
		x += k * xp[0];
		y += k * yp[0];

		for (tjs_int i = 1; i <= N-1; i++) {
			tjs_real c = combination[i];
			tjs_real k = c * pow(t, i) * pow(s, N-i);
			x += k * xp[i];
			y += k * yp[i];
		}

		k = pow(t, N);
		x += k * xp[N];
		y += k * yp[N];

		rv = true;
	}

	cx = x;
	cy = y;

	return rv;
}






//--------------------------------
// Lagrange Interpolator
//--------------------------------

void LagrangeCurveTracer::makeTable()
{
	allocMem();

	tjs_int * x = this->xpath;
	tjs_int * y = this->ypath;
	tjs_real * p = this->pz;
	tjs_int N = this->numPoints;

	p[0] = 0;

	for (tjs_int i = 1; i < N; i++) {
		tjs_int t1 = x[i] - x[i - 1];
		tjs_int t2 = y[i] - y[i - 1];
		p[i] = p[i - 1] + sqrt(t1*t1 + t2*t2);
	}

	for (tjs_int i = 1; i < N; i++) {
		p[i] /= p[N - 1];
	}

//	createNodeIndex();

	return;
}

bool LagrangeCurveTracer::getCoord(tjs_int dindex, tjs_real & cx, tjs_real & cy)
{
	bool rv = false;
	if (0 <= dindex && dindex <= this->numNodes) {
		cx = internalGetCoord(dindex, this->xpath);
		cy = internalGetCoord(dindex, this->ypath);
		rv = true;
	}
	return rv;
}

tjs_real LagrangeCurveTracer::internalGetCoord(tjs_int index, tjs_int * y)
{
	tjs_real rv = 0;
	tjs_real x = (tjs_real)index / (tjs_real)numNodes;
	tjs_real * p = this->pz;
	tjs_int N = this->numPoints;

	for (tjs_int i = 0; i < N; i++) {
		tjs_real xi = p[i];
		tjs_real t = 1;
		for (tjs_int j = 0; j < N; j++) {
			if (i == j) continue;
			tjs_real xj = p[j];
			tjs_real d = (x - xj) / (xi - xj);
			t *= d;
		}
		rv += (tjs_real)(y[i]) * t;
	}

	return rv;
}




//--------------------------------
// Chaikin (open)
//--------------------------------
void ChaikinTracerOpen::init(tjs_int * xp, tjs_int * yp, tjs_int n, tjs_int ac)
{
	clear();
	if (n >= getRequiredPoints() && ac > 0 && xp && yp) {
		this->numPoints = n;
		this->xpath = xp;
		this->ypath = yp;
		this->accuracy = ac;
		this->numNodes = (numPoints - 1) * accuracy;
		makeTable();
		activate();
	}

	return;
}

void ChaikinTracerOpen::clear()
{
	xpath = ypath = NULL;
	inactivate();
}


void ChaikinTracerOpen::makeTable()
{
	allocMem();

	tjs_int * x = this->xpath;
	tjs_int * y = this->ypath;
	tjs_int * x2 = this->xpath2;
	tjs_int * y2 = this->ypath2;
	tjs_int N  = this->numPoints;
	tjs_int N2 = this->numPoints2;

	// ႤƂ͂ --->

	// n_
	x2[0] = (x[1] + x[0]) / 2;
	y2[0] = (y[1] + y[0]) / 2;

	// N2 = 2*N-3
	// _W
	for (tjs_int i = 1; i < N - 1; i++) {
		x2[2*i-1] = x[i];
		y2[2*i-1] = y[i];
		x2[2*i] = (x[i + 1] + x[i]) / 2;
		y2[2*i] = (y[i + 1] + y[i]) / 2;
	}

	// <--- ܂

	tjs_real * p = this->pz;

	p[0] = 0;

	for (tjs_int i = 1; i < N2; i++) {
		tjs_real t1 = x2[i] - x2[i - 1];
		tjs_real t2 = y2[i] - y2[i - 1];
		p[i] = p[i - 1] + sqrt(t1*t1 + t2*t2);
	}

	for (tjs_int i = 1; i < N2; i++) {
		p[i] /= p[N2-1];
	}

	createNodeIndex();
}


bool ChaikinTracerOpen::getCoord(tjs_int dindex, tjs_real & cx, tjs_real & cy)
{
	if (dindex < 0 || dindex > numNodes ) return false;

	tjs_real * p = this->pz;
	tjs_int i = this->nodeIndex[dindex];

	if (i > 0) i--;
	if (i % 2) i--; // ɂ

	tjs_real t = (tjs_real)dindex / (tjs_real)numNodes;
	t = (t - p[i]) / (p[i + 2] - p[i]);
	tjs_real s = 1.0 - t;
	tjs_real tt = t * t;
	tjs_real ts = t * s;
	tjs_real ss = s * s;
	tjs_int * x2 = this->xpath2;
	tjs_int * y2 = this->ypath2;

	cx = ((tjs_real)x2[i] * ss) + (2.0 * (tjs_real)x2[i + 1] * ts) + ((tjs_real)x2[i + 2] * tt);
	cy = ((tjs_real)y2[i] * ss) + (2.0 * (tjs_real)y2[i + 1] * ts) + ((tjs_real)y2[i + 2] * tt);

	return true;
}


void ChaikinTracerOpen::createNodeIndex()
{
	tjs_real * p = this->pz;
	tjs_int * ni = this->nodeIndex;
	tjs_int N2 = this->numPoints2;
	tjs_int ac = this->numNodes;

	tjs_int k = 0;

	for (tjs_int i = 0; i <= ac; i++) {
		tjs_real t = (tjs_real)i / (tjs_real)ac;
		if (t > p[k] && (k < N2 - 1)) {
			k++;
		}
		ni[i] = k;
	}
}

void ChaikinTracerOpen::releaseMem()
{
	DELETE_ARRAY(xpath2)
	DELETE_ARRAY(ypath2)
	DELETE_ARRAY(pz)
	DELETE_ARRAY(nodeIndex)
}

void ChaikinTracerOpen::allocMem()
{
	releaseMem();
	numPoints2 = 2 * numPoints - 3;
	this->pz = new tjs_real[numPoints2];
	this->xpath2 = new tjs_int[numPoints2];
	this->ypath2 = new tjs_int[numPoints2];
	this->nodeIndex = new tjs_int[numNodes + 1];
}



//--------------------------------
// Chaikin (open) PST
//--------------------------------
void ChaikinTracerOpenPST::makeTable()
{
	allocMem();

	tjs_int * x = this->xpath;
	tjs_int * y = this->ypath;
	tjs_int * x2 = this->xpath2;
	tjs_int * y2 = this->ypath2;
	tjs_int N  = this->numPoints;
	tjs_int N2 = this->numPoints2;

	// ႤƂ͂ --->

	// n_ƏI_
	x2[0] = x[0];
	y2[0] = y[0];
	x2[N2 - 2] = x[N - 2];
	y2[N2 - 2] = y[N - 2];
	x2[N2 - 1] = x[N - 1];
	y2[N2 - 1] = y[N - 1];
	// _W
	for (tjs_int i = 1; i < N - 2; i++) {
		x2[2*i-1] = x[i];
		y2[2*i-1] = y[i];
		x2[2*i] = (x[i + 1] + x[i]) / 2;
		y2[2*i] = (y[i + 1] + y[i]) / 2;
	}

	// <--- ܂

	tjs_real * p = this->pz;

	p[0] = 0;

	for (tjs_int i = 1; i < N2; i++) {
		tjs_real t1 = x2[i] - x2[i - 1];
		tjs_real t2 = y2[i] - y2[i - 1];
		p[i] = p[i - 1] + sqrt(t1*t1 + t2*t2);
	}

	for (tjs_int i = 1; i < N2; i++) {
		p[i] /= p[N2-1];
	}

	createNodeIndex();
}





//--------------------------------
// Chaikin (close)
//--------------------------------
void ChaikinTracerClose::init(tjs_int * xp, tjs_int * yp, tjs_int n, tjs_int ac)
{
	clear();
	if (n >= getRequiredPoints() && ac > 0 && xp && yp && (xp[0] == xp[n-1]) && (yp[0] == yp[n-1])) {
		this->numPoints = n;
		this->xpath = xp;
		this->ypath = yp;
		this->accuracy = ac;
		this->numNodes = (numPoints - 1) * accuracy;
		makeTable();
		activate();
	}

	return;
}


void ChaikinTracerClose::makeTable()
{
	allocMem();

	tjs_int * x = this->xpath;
	tjs_int * y = this->ypath;
	tjs_int * x2 = this->xpath2;
	tjs_int * y2 = this->ypath2;
	tjs_int N  = this->numPoints;
	tjs_int N2 = this->numPoints2;

	// _W
	for (tjs_int i = 0; i < N - 1; i++) {
		tjs_int k = i + 1;
		x2[2*i] = (x[k] + x[i]) / 2;
		y2[2*i] = (y[k] + y[i]) / 2;
		x2[2*i+1] = x[k];
		y2[2*i+1] = y[k];
	}
	x2[N2 - 1] = x2[0];
	y2[N2 - 1] = y2[0];

	tjs_real * p = this->pz;

	p[0] = 0;

	for (tjs_int i = 1; i < N2; i++) {
		tjs_real t1 = x2[i] - x2[i - 1];
		tjs_real t2 = y2[i] - y2[i - 1];
		p[i] = p[i - 1] + sqrt(t1*t1 + t2*t2);
	}

	for (tjs_int i = 1; i < N2; i++) {
		p[i] /= p[N2-1];
	}

	createNodeIndex();
}

void ChaikinTracerClose::allocMem()
{
	releaseMem();
	numPoints2 = 2 * numPoints - 1;
	this->pz = new tjs_real[numPoints2];
	this->xpath2 = new tjs_int[numPoints2];
	this->ypath2 = new tjs_int[numPoints2];
	this->nodeIndex = new tjs_int[numNodes + 1];
}


